@@ -29,12 +29,10 @@ using namespace std;
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <sys/types.h>
 #include <utime.h>
 #include <math.h>
 #include <sys/statvfs.h>
 
-#include "common/config.h"
 #include "include/assert.h"
 
 #define dout_subsys ceph_subsys_client
@@ -318,9 +316,15 @@ string SyntheticClient::get_sarg(int seq)
 int SyntheticClient::run()
 { 
   dout(15) << "initing" << dendl;
-  client->init();
+  int err = client->init();
+  if (err < 0) {
+    char buf[80];
+    dout(0) << "failed to initialize: " << strerror_r(-err, buf, sizeof(buf)) << dendl;
+    return -1;
+  }
+
   dout(15) << "mounting" << dendl;
-  int err = client->mount("");
+  err = client->mount("");
   if (err < 0) {
     char buf[80];
     dout(0) << "failed to mount: " << strerror_r(-err, buf, sizeof(buf)) << dendl;
@@ -1678,11 +1682,10 @@ int SyntheticClient::dump_placement(string& fn) {
     int osd = client->osdmap->get_pg_primary(client->osdmap->object_locator_to_pg(i->oid, i->oloc));
 
     // run through all the buffer extents
-    for (map<uint64_t, uint64_t>::iterator j = i ->buffer_extents.begin();
+    for (vector<pair<uint64_t, uint64_t> >::iterator j = i->buffer_extents.begin();
 	 j != i->buffer_extents.end(); ++j) {
-      
-      dout(0) << "OSD " << osd << ", offset " << (*j).first <<
-	", length " << (*j).second << dendl;    
+      dout(0) << "OSD " << osd << ", offset " << (*j).first
+	      << ", length " << (*j).second << dendl;    
     }
   }
   return 0;
@@ -1974,7 +1977,10 @@ int SyntheticClient::write_file(string& fn, int size, loff_t wrsize)   // size i
 
   int fd = client->open(fn.c_str(), O_RDWR|O_CREAT);
   dout(5) << "writing to " << fn << " fd " << fd << dendl;
-  if (fd < 0) return fd;
+  if (fd < 0) {
+    delete[] buf;
+    return fd;
+  }
   
   utime_t from = ceph_clock_now(g_ceph_context);
   utime_t start = from;
@@ -2034,7 +2040,10 @@ int SyntheticClient::write_fd(int fd, int size, int wrsize)   // size is in MB,
   uint64_t chunks = (uint64_t)size * (uint64_t)(1024*1024) / (uint64_t)wrsize;
 
   //dout(5) << "SyntheticClient::write_fd: writing to fd " << fd << dendl;
-  if (fd < 0) return fd;
+  if (fd < 0) {
+    delete[] buf;
+    return fd;
+  }
 
   for (unsigned i=0; i<chunks; i++) {
     if (time_to_stop()) {
@@ -2084,7 +2093,10 @@ int SyntheticClient::read_file(const std::string& fn, int size,
 
   int fd = client->open(fn.c_str(), O_RDONLY);
   dout(5) << "reading from " << fn << " fd " << fd << dendl;
-  if (fd < 0) return fd;
+  if (fd < 0) {
+    delete[] buf;
+    return fd;
+  }
 
   utime_t from = ceph_clock_now(g_ceph_context);
   utime_t start = from;
@@ -2351,9 +2363,9 @@ int SyntheticClient::object_rw(int nobj, int osize, int wrpc,
     lat -= start;
     if (client->logger) {
       if (write) 
-	client->logger->fset(l_c_owrlat, lat);
+	client->logger->tset(l_c_owrlat, lat);
       else 
-	client->logger->fset(l_c_ordlat, lat);
+	client->logger->tset(l_c_ordlat, lat);
     }
   }
 
