@@ -97,6 +97,28 @@ public:
   }
 
   // tunables
+  void set_tunables_legacy() {
+    crush->choose_local_tries = 2;
+    crush->choose_local_fallback_tries = 5;
+    crush->choose_total_tries = 19;
+    crush->chooseleaf_descend_once = 0;
+  }
+  void set_tunables_optimal() {
+    crush->choose_local_tries = 0;
+    crush->choose_local_fallback_tries = 0;
+    crush->choose_total_tries = 50;
+    crush->chooseleaf_descend_once = 1;
+  }
+  void set_tunables_argonaut() {
+    set_tunables_legacy();
+  }
+  void set_tunables_bobtail() {
+    set_tunables_optimal();
+  }
+  void set_tunables_default() {
+    set_tunables_legacy();
+  }
+
   int get_choose_local_tries() const {
     return crush->choose_local_tries;
   }
