@@ -67,7 +67,6 @@ public:
     string symlink;
     bufferlist snapbl;
     bool dirty;
-    struct default_file_layout *dir_layout;
     typedef map<snapid_t, old_inode_t> old_inodes_t;
     old_inodes_t old_inodes;
 
@@ -79,11 +78,11 @@ public:
     fullbit(const string& d, snapid_t df, snapid_t dl, 
 	    version_t v, inode_t& i, fragtree_t &dft, 
 	    map<string,bufferptr> &xa, const string& sym,
-	    bufferlist &sbl, bool dr, default_file_layout *defl = NULL,
+	    bufferlist &sbl, bool dr,
 	    old_inodes_t *oi = NULL) :
       //dn(d), dnfirst(df), dnlast(dl), dnv(v), 
       //inode(i), dirfragtree(dft), xattrs(xa), symlink(sym), snapbl(sbl), dirty(dr) 
-      dir_layout(NULL), _enc(1024)
+      _enc(1024)
     {
       ::encode(d, _enc);
       ::encode(df, _enc);
@@ -96,25 +95,20 @@ public:
       if (i.is_dir()) {
 	::encode(dft, _enc);
 	::encode(sbl, _enc);
-	::encode((defl ? true : false), _enc);
-	if (defl)
-	  ::encode(*defl, _enc);
       }
       ::encode(dr, _enc);      
       ::encode(oi ? true : false, _enc);
       if (oi)
 	::encode(*oi, _enc);
     }
-    fullbit(bufferlist::iterator &p) : dir_layout(NULL) {
+    fullbit(bufferlist::iterator &p) {
       decode(p);
     }
-    fullbit() : dir_layout(NULL) {}
-    ~fullbit() {
-      delete dir_layout;
-    }
+    fullbit() {}
+    ~fullbit() {}
 
     void encode(bufferlist& bl) const {
-      __u8 struct_v = 3;
+      __u8 struct_v = 4;
       ::encode(struct_v, bl);
       assert(_enc.length());
       bl.append(_enc); 
@@ -133,12 +127,13 @@ public:
       if (inode.is_dir()) {
 	::decode(dirfragtree, bl);
 	::decode(snapbl, bl);
-	if (struct_v >= 2) {
+	if ((struct_v == 2) || (struct_v == 3)) {
 	  bool dir_layout_exists;
 	  ::decode(dir_layout_exists, bl);
 	  if (dir_layout_exists) {
-	    dir_layout = new default_file_layout;
-	    ::decode(*dir_layout, bl);
+	    __u8 dir_struct_v;
+	    ::decode(dir_struct_v, bl); // default_file_layout version
+	    ::decode(inode.layout, bl); // and actual layout, that we care about
 	  }
 	}
       }
@@ -559,11 +554,6 @@ private:
     //cout << "journaling " << in->inode.ino << " at " << my_offset << std::endl;
 
     inode_t *pi = in->get_projected_inode();
-    default_file_layout *default_layout = NULL;
-    if (in->is_dir())
-      default_layout = (in->get_projected_node() ?
-                           in->get_projected_node()->dir_layout :
-                           in->default_layout);
 
     bufferlist snapbl;
     sr_t *sr = in->get_projected_srnode();
@@ -577,7 +567,7 @@ private:
 									 *pi, in->dirfragtree,
 									 *in->get_projected_xattrs(),
 									 in->symlink, snapbl,
-									 dirty, default_layout,
+									 dirty,
 									 &in->old_inodes)));
   }
 
@@ -608,12 +598,6 @@ private:
     if (!pdft) pdft = &in->dirfragtree;
     if (!px) px = &in->xattrs;
 
-    default_file_layout *default_layout = NULL;
-    if (in->is_dir())
-      default_layout = (in->get_projected_node() ?
-                           in->get_projected_node()->dir_layout :
-                           in->default_layout);
-
     bufferlist snapbl;
     if (psnapbl)
       snapbl = *psnapbl;
@@ -627,7 +611,7 @@ private:
 		       0,
 		       *pi, *pdft, *px,
 		       in->symlink, snapbl,
-		       dirty, default_layout, &in->old_inodes);
+		       dirty, &in->old_inodes);
   }
   
   dirlump& add_dir(CDir *dir, bool dirty, bool complete=false, bool isnew=false) {
