@@ -49,9 +49,12 @@ MonClient::MonClient(CephContext *cct_) :
   state(MC_STATE_NONE),
   messenger(NULL),
   cur_con(NULL),
+  rng(getpid()),
   monc_lock("MonClient::monc_lock"),
   timer(cct_, monc_lock), finisher(cct_),
+  authorize_handler_registry(NULL),
   initialized(false),
+  no_keyring_disabled_cephx(false),
   log_client(NULL),
   more_log_pending(false),
   auth_supported(NULL),
@@ -118,7 +121,7 @@ int MonClient::get_monmap_privately()
   ldout(cct, 10) << "have " << monmap.epoch << " fsid " << monmap.fsid << dendl;
   
   while (monmap.fsid.is_zero()) {
-    cur_mon = monmap.pick_random_mon();
+    cur_mon = _pick_random_mon();
     cur_con = messenger->get_connection(monmap.get_inst(cur_mon));
     ldout(cct, 10) << "querying mon." << cur_mon << " " << cur_con->get_peer_addr() << dendl;
     messenger->send_message(new MMonGetMap, cur_con);
@@ -126,7 +129,8 @@ int MonClient::get_monmap_privately()
     if (--attempt == 0)
       break;
     
-    utime_t interval(1, 0);
+    utime_t interval;
+    interval.set_from_double(cct->_conf->mon_client_hunt_interval);
     map_cond.WaitInterval(cct, monc_lock, interval);
 
     if (monmap.fsid.is_zero()) {
@@ -261,50 +265,48 @@ int MonClient::init()
 
   entity_name = cct->_conf->name;
 
-  // keyring
-  keyring = new KeyRing;
-  int r = keyring->from_ceph_context(cct);
-  if (r == -ENOENT) {
-    // do we care?
-    string method;
-    if (entity_name.get_type() == CEPH_ENTITY_TYPE_MDS ||
-	entity_name.get_type() == CEPH_ENTITY_TYPE_OSD)
+  Mutex::Locker l(monc_lock);
+
+  string method;
+    if (cct->_conf->auth_supported.length() != 0)
+      method = cct->_conf->auth_supported;
+    else if (entity_name.get_type() == CEPH_ENTITY_TYPE_OSD ||
+             entity_name.get_type() == CEPH_ENTITY_TYPE_MDS ||
+             entity_name.get_type() == CEPH_ENTITY_TYPE_MON)
       method = cct->_conf->auth_cluster_required;
     else
       method = cct->_conf->auth_client_required;
-    if (method.length() == 0)
-      method = cct->_conf->auth_supported;
-    AuthMethodList supported(cct, method);
-    if (!supported.is_supported_auth(CEPH_AUTH_CEPHX)) {
-      ldout(cct, 2) << "cephx auth is not supported, ignoring absence of keyring" << dendl;
-      r = 0;
+  auth_supported = new AuthMethodList(cct, method);
+  ldout(cct, 10) << "auth_supported " << auth_supported->get_supported_set() << " method " << method << dendl;
+
+  int r = 0;
+  keyring = new KeyRing; // initializing keyring anyway
+
+  if (auth_supported->is_supported_auth(CEPH_AUTH_CEPHX)) {
+    r = keyring->from_ceph_context(cct);
+    if (r == -ENOENT) {
+      auth_supported->remove_supported_auth(CEPH_AUTH_CEPHX);
+      if (auth_supported->get_supported_set().size() > 0) {
+	r = 0;
+	no_keyring_disabled_cephx = true;
+      } else {
+	lderr(cct) << "ERROR: missing keyring, cannot use cephx for authentication" << dendl;
+      }
     }
   }
+
   if (r < 0) {
-    lderr(cct) << "failed to open keyring: " << cpp_strerror(r) << dendl;
     return r;
   }
+
   rotating_secrets = new RotatingKeyRing(cct, cct->get_module_type(), keyring);
 
-  Mutex::Locker l(monc_lock);
+  initialized = true;
+
   timer.init();
   finisher.start();
   schedule_tick();
 
-  // seed rng so we choose a different monitor each time
-  srand(getpid());
-
-  string method;
-  if (entity_name.get_type() == CEPH_ENTITY_TYPE_OSD ||
-      entity_name.get_type() == CEPH_ENTITY_TYPE_MDS ||
-      entity_name.get_type() == CEPH_ENTITY_TYPE_MON)
-    method = cct->_conf->auth_cluster_required;
-  else
-    method = cct->_conf->auth_client_required;
-  auth_supported = new AuthMethodList(cct, method.length() ? method : cct->_conf->auth_supported);
-  ldout(cct, 10) << "auth_supported " << auth_supported->get_supported_set() << dendl;
-
-  initialized = true;
   return 0;
 }
 
@@ -356,6 +358,10 @@ int MonClient::authenticate(double timeout)
     ldout(cct, 5) << "authenticate success, global_id " << global_id << dendl;
   }
 
+  if (authenticate_err < 0 && no_keyring_disabled_cephx) {
+    lderr(cct) << "authenticate NOTE: no keyring found; disabled cephx authentication" << dendl;
+  }
+
   return authenticate_err;
 }
 
@@ -442,24 +448,39 @@ void MonClient::_send_mon_message(Message *m, bool force)
   }
 }
 
+string MonClient::_pick_random_mon()
+{
+  assert(monmap.size() > 0);
+  if (monmap.size() == 1) {
+    return monmap.get_name(0);
+  } else {
+    int max = monmap.size();
+    int o = -1;
+    if (!cur_mon.empty()) {
+      o = monmap.get_rank(cur_mon);
+      if (o >= 0)
+	max--;
+    }
+
+    int32_t n = rng() % max;
+    if (o >= 0 && n >= o)
+      n++;
+    return monmap.get_name(n);
+  }
+}
+
 void MonClient::_pick_new_mon()
 {
   assert(monc_lock.is_locked());
 
-  assert(monmap.size() > 0);
-  if (!cur_mon.empty() && monmap.size() > 1) {
-    // pick a _different_ mon
-    cur_mon = monmap.pick_random_mon_not(cur_mon);
-  } else {
-    cur_mon = monmap.pick_random_mon();
-  }
+  cur_mon = _pick_random_mon();
 
   if (cur_con) {
     messenger->mark_down(cur_con);
     cur_con->put();
   }
   cur_con = messenger->get_connection(monmap.get_inst(cur_mon));
-
+	
   ldout(cct, 10) << "_pick_new_mon picked mon." << cur_mon << " con " << cur_con
 		 << " addr " << cur_con->get_peer_addr()
 		 << dendl;
@@ -690,23 +711,6 @@ int MonClient::wait_auth_rotating(double timeout)
 
 // ---------
 
-struct C_IsLatestMap : public Context {
-  Context *onfinish;
-  version_t newest;
-  version_t have;
-  C_IsLatestMap(Context *f, version_t h) : onfinish(f), newest(0), have(h) {}
-  void finish(int r) {
-    onfinish->complete(r == 0 ? (have != newest) : r);
-  }
-};
-
-void MonClient::is_latest_map(string map, version_t cur_ver, Context *onfinish)
-{
-  ldout(cct, 10) << "is_latest_map " << map << " current " << cur_ver << dendl;;
-  C_IsLatestMap *c = new C_IsLatestMap(onfinish, cur_ver);
-  get_version(map, &c->newest, NULL, c);
-}
-
 void MonClient::get_version(string map, version_t *newest, version_t *oldest, Context *onfinish)
 {
   ldout(cct, 10) << "get_version " << map << dendl;
