@@ -27,8 +27,9 @@
 
 #include <sstream>
 #include "common/config.h"
+#include "include/assert.h"
 
-#define DOUT_SUBSYS mon
+#define dout_subsys ceph_subsys_mon
 #undef dout_prefix
 #define dout_prefix _prefix(_dout, mon)
 static ostream& _prefix(std::ostream *_dout, Monitor *mon) {
@@ -39,11 +40,8 @@ static ostream& _prefix(std::ostream *_dout, Monitor *mon) {
 
 void MonmapMonitor::create_initial()
 {
-  bufferlist bl;
-  mon->store->get_bl_ss(bl, "mkfs", "monmap");
-  pending_map.decode(bl);
-  dout(10) << "create_initial set fed epoch " << pending_map.epoch << dendl;
-  assert(pending_map.epoch == 0);   // fix mkfs()
+  dout(10) << "create_initial using current monmap" << dendl;
+  pending_map = *mon->monmap;
   pending_map.epoch = 1;
 }
 
@@ -102,7 +100,20 @@ void MonmapMonitor::encode_pending(bufferlist& bl)
 
   assert(mon->monmap->epoch + 1 == pending_map.epoch ||
 	 pending_map.epoch == 1);  // special case mkfs!
-  pending_map.encode(bl);
+  pending_map.encode(bl, mon->get_quorum_features());
+}
+
+void MonmapMonitor::on_active()
+{
+  if (paxos->get_version() >= 1 && !mon->has_ever_joined) {
+    // make note of the fact that i was, once, part of the quorum.
+    dout(10) << "noting that i was, once, part of an active quorum." << dendl;
+    mon->store->put_int(1, "joined");
+    mon->has_ever_joined = true;
+  }
+
+  if (mon->is_leader())
+    mon->clog.info() << "monmap " << *mon->monmap << "\n";
 }
 
 bool MonmapMonitor::preprocess_query(PaxosServiceMessage *m)
@@ -120,12 +131,32 @@ bool MonmapMonitor::preprocess_query(PaxosServiceMessage *m)
   }
 }
 
+void MonmapMonitor::dump_info(Formatter *f)
+{
+  f->open_object_section("monmap");
+  mon->monmap->dump(f);
+  f->open_array_section("quorum");
+  for (set<int>::iterator q = mon->get_quorum().begin(); q != mon->get_quorum().end(); ++q)
+    f->dump_int("mon", *q);
+  f->close_section();
+  f->close_section();
+}
+
 bool MonmapMonitor::preprocess_command(MMonCommand *m)
 {
   int r = -1;
   bufferlist rdata;
   stringstream ss;
 
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_MONMAP, MON_CAP_R) &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", paxos->get_version());
+    return true;
+  }
+
   vector<const char*> args;
   for (unsigned i = 1; i < m->cmd.size(); i++)
     args.push_back(m->cmd[i].c_str());
@@ -133,11 +164,12 @@ bool MonmapMonitor::preprocess_command(MMonCommand *m)
   if (m->cmd.size() > 1) {
     if (m->cmd[1] == "stat") {
       mon->monmap->print_summary(ss);
-      ss << ", election epoch " << mon->get_epoch() << ", quorum " << mon->get_quorum();
+      ss << ", election epoch " << mon->get_epoch() << ", quorum " << mon->get_quorum()
+	 << " " << mon->get_quorum_names();
       r = 0;
     }
     else if (m->cmd.size() == 2 && m->cmd[1] == "getmap") {
-      mon->monmap->encode(rdata);
+      mon->monmap->encode(rdata, CEPH_FEATURES_ALL);
       r = 0;
       ss << "got latest monmap";
     }
@@ -149,11 +181,17 @@ bool MonmapMonitor::preprocess_command(MMonCommand *m)
       for (std::vector<const char*>::iterator i = args.begin()+1; i != args.end(); ) {
 	if (ceph_argparse_double_dash(args, i))
 	  break;
-	else if (ceph_argparse_witharg(args, i, &val, "-f", "--format", (char*)NULL))
+	else if (ceph_argparse_witharg_daemon(args, i, &val, "-f", "--format",
+					      (char*)NULL))
 	  format = val;
-	else if (!epoch)
-	  epoch = atoi(*i++);
-	else
+	else if (!epoch) {
+	  long l = parse_pos_long(*i++, &ss);
+	  if (l < 0) {
+	    r = -EINVAL;
+	    goto out;
+	  }
+	  epoch = l;
+	} else
 	  i++;
       }
 
@@ -215,18 +253,24 @@ bool MonmapMonitor::preprocess_command(MMonCommand *m)
         r = 0;
       } else {
         // find target
-        int target = atoi(m->cmd[2].c_str());
-        stringstream ss;
-        if (target == 0 && m->cmd[2] != "0") {
-          ss << "could not parse target " << m->cmd[2];
-        } else {
-          // send to target, or handle if it's me
-   	  MMonCommand *newm = new MMonCommand(m->fsid, m->version);
-	  newm->cmd.insert(newm->cmd.begin(), m->cmd.begin() + 3, m->cmd.end());
-	  mon->messenger->send_message(newm, mon->monmap->get_inst(target));
-	  ss << "fw to mon." << target;
-	  r = 0;
-        }
+        long target = parse_pos_long(m->cmd[2].c_str(), &ss);
+	if (target < 0) {
+	  r = -EINVAL;
+	  goto out;
+	}
+	if (target >= (long)mon->monmap->size()) {
+	  ss << "mon." << target << " does not exist";
+	  r = -ENOENT;
+	  goto out;
+	}
+
+	// send to target, or handle if it's me
+	stringstream ss;
+	MMonCommand *newm = new MMonCommand(m->fsid, m->version);
+	newm->cmd.insert(newm->cmd.begin(), m->cmd.begin() + 3, m->cmd.end());
+	mon->messenger->send_message(newm, mon->monmap->get_inst(target));
+	ss << "fw to mon." << target;
+	r = 0;
       }
     }
     else if (m->cmd[1] == "add")
@@ -235,6 +279,7 @@ bool MonmapMonitor::preprocess_command(MMonCommand *m)
       return false;
   }
 
+ out:
   if (r != -1) {
     string rs;
     getline(ss, rs);
@@ -268,6 +313,16 @@ bool MonmapMonitor::prepare_command(MMonCommand *m)
   stringstream ss;
   string rs;
   int err = -EINVAL;
+
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_MONMAP, MON_CAP_R) &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", paxos->get_version());
+    return true;
+  }
+
   if (m->cmd.size() > 1) {
     if (m->cmd.size() == 4 && m->cmd[1] == "add") {
       string name = m->cmd[2];
@@ -279,9 +334,17 @@ bool MonmapMonitor::prepare_command(MMonCommand *m)
 	ss << "addr " << m->cmd[3] << "does not parse";
 	goto out;
       }
+
+      if (addr.get_port() == 0) {
+	ss << "port defaulted to " << CEPH_MON_PORT;
+	addr.set_port(CEPH_MON_PORT);
+      }
+
       if (pending_map.contains(addr) ||
 	  pending_map.contains(name)) {
 	err = -EEXIST;
+	if (!ss.str().empty())
+	  ss << "; ";
 	ss << "mon " << name << " " << addr << " already exists";
 	goto out;
       }
@@ -325,12 +388,21 @@ bool MonmapMonitor::preprocess_join(MMonJoin *join)
 {
   dout(10) << "preprocess_join " << join->name << " at " << join->addr << dendl;
 
-  if (pending_map.contains(join->name)) {
+  MonSession *session = join->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_MONMAP, MON_CAP_ALL))) {
+    dout(10) << " insufficient caps" << dendl;
+    join->put();
+    return true;
+  }
+
+  if (pending_map.contains(join->name) && !pending_map.get_addr(join->name).is_blank_ip()) {
     dout(10) << " already have " << join->name << dendl;
     join->put();
     return true;
   }
-  if (pending_map.contains(join->addr)) {
+  if (pending_map.contains(join->addr) && pending_map.get_name(join->addr) == join->name) {
     dout(10) << " already have " << join->addr << dendl;
     join->put();
     return true;
@@ -339,7 +411,11 @@ bool MonmapMonitor::preprocess_join(MMonJoin *join)
 }
 bool MonmapMonitor::prepare_join(MMonJoin *join)
 {
-  dout(0) << "adding " << join->name << " at " << join->addr << " to monitor cluster" << dendl;
+  dout(0) << "adding/updating " << join->name << " at " << join->addr << " to monitor cluster" << dendl;
+  if (pending_map.contains(join->name))
+    pending_map.remove(join->name);
+  if (pending_map.contains(join->addr))
+    pending_map.remove(pending_map.get_name(join->addr));
   pending_map.add(join->name, join->addr);
   pending_map.last_changed = ceph_clock_now(g_ceph_context);
   join->put();
@@ -357,16 +433,26 @@ void MonmapMonitor::tick()
   update_from_paxos();
 }
 
-enum health_status_t MonmapMonitor::get_health(std::ostream &ss) const
+void MonmapMonitor::get_health(list<pair<health_status_t, string> >& summary,
+			       list<pair<health_status_t, string> > *detail) const
 {
-  enum health_status_t ret(HEALTH_OK);
-  
   int max = mon->monmap->size();
   int actual = mon->get_quorum().size();
   if (actual < max) {
-    ret = HEALTH_WARN;
-    ss << (max-actual) << " mons down, quorum " << mon->get_quorum();
+    ostringstream ss;
+    ss << (max-actual) << " mons down, quorum " << mon->get_quorum() << " " << mon->get_quorum_names();
+    summary.push_back(make_pair(HEALTH_WARN, ss.str()));
+    if (detail) {
+      set<int> q = mon->get_quorum();
+      for (int i=0; i<max; i++) {
+	if (q.count(i) == 0) {
+	  ostringstream ss;
+	  ss << "mon." << mon->monmap->get_name(i) << " (rank " << i
+	     << ") addr " << mon->monmap->get_addr(i)
+	     << " is down (out of quorum)";
+	  detail->push_back(make_pair(HEALTH_WARN, ss.str()));
+	}
+      }
+    }
   }
-
-  return ret;
 }
