@@ -35,7 +35,7 @@ public:
     
   void *entry();
   void stop();
-  int bind(entity_addr_t &bind_addr, int avoid_port1=0, int avoid_port2=0);
+  int bind(const entity_addr_t &bind_addr, int avoid_port1=0, int avoid_port2=0);
   int rebind(int avoid_port);
   int start();
 };
