@@ -380,6 +380,10 @@ Connection *SimpleMessenger::get_connection(const entity_inst_t& dest)
   }
 }
 
+Connection *SimpleMessenger::get_loopback_connection()
+{
+  return (Connection*)local_connection->get();
+}
 
 void SimpleMessenger::submit_message(Message *m, Connection *con,
 				     const entity_addr_t& dest_addr, int dest_type, bool lazy)
@@ -407,6 +411,8 @@ void SimpleMessenger::submit_message(Message *m, Connection *con,
       pipe->put();
       ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr
 		    << ", had pipe " << pipe << ", but it closed." << dendl;
+      m->put();
+      return;
     }
   }
 
