@@ -49,6 +49,8 @@ int cls_register_method(cls_handle_t hclass, const char *method,
                         int flags,
                         cls_method_call_t class_call, cls_method_handle_t *handle)
 {
+  if (!(flags & (CLS_METHOD_RD | CLS_METHOD_WR)))
+    return -EINVAL;
   ClassHandler::ClassData *cls = (ClassHandler::ClassData *)hclass;
   cls_method_handle_t hmethod =(cls_method_handle_t)cls->register_method(method, flags, class_call);
   if (handle)
@@ -182,6 +184,14 @@ int cls_cxx_create(cls_method_context_t hctx, bool exclusive)
   return (*pctx)->pg->do_osd_ops(*pctx, ops);
 }
 
+int cls_cxx_remove(cls_method_context_t hctx)
+{
+  ReplicatedPG::OpContext **pctx = (ReplicatedPG::OpContext **)hctx;
+  vector<OSDOp> ops(1);
+  ops[0].op.op = CEPH_OSD_OP_DELETE;
+  return (*pctx)->pg->do_osd_ops(*pctx, ops);
+}
+
 int cls_cxx_stat(cls_method_context_t hctx, uint64_t *size, time_t *mtime)
 {
   ReplicatedPG::OpContext **pctx = (ReplicatedPG::OpContext **)hctx;
@@ -480,7 +490,7 @@ int cls_cxx_map_set_val(cls_method_context_t hctx, const string &key,
 }
 
 int cls_cxx_map_set_vals(cls_method_context_t hctx,
-			 std::map<string, bufferlist> *map)
+			 const std::map<string, bufferlist> *map)
 {
   ReplicatedPG::OpContext **pctx = (ReplicatedPG::OpContext **)hctx;
   vector<OSDOp> ops(1);
