@@ -149,7 +149,6 @@ public:
     on_deletion_complete.push_front(completion);
   }
   ~DeletingState() {
-    Mutex::Locker l(lock);
     for (list<Context *>::iterator i = on_deletion_complete.begin();
 	 i != on_deletion_complete.end();
 	 ++i) {
@@ -617,6 +616,11 @@ private:
     finished.splice(finished.end(), ls);
     finished_lock.Unlock();
   }
+  void take_waiters_front(list<OpRequestRef>& ls) {
+    finished_lock.Lock();
+    finished.splice(finished.begin(), ls);
+    finished_lock.Unlock();
+  }
   void take_waiter(OpRequestRef op) {
     finished_lock.Lock();
     finished.push_back(op);
@@ -786,7 +790,7 @@ private:
   epoch_t note_peer_epoch(int p, epoch_t e);
   void forget_peer_epoch(int p, epoch_t e);
 
-  bool _share_map_incoming(const entity_inst_t& inst, epoch_t epoch,
+  bool _share_map_incoming(entity_name_t name, Connection *con, epoch_t epoch,
 			   Session *session = 0);
   void _share_map_outgoing(int peer, Connection *con,
 			   OSDMapRef map = OSDMapRef());
@@ -833,9 +837,7 @@ private:
   }
 
   MOSDMap *build_incremental_map_msg(epoch_t from, epoch_t to);
-  void send_incremental_map(epoch_t since, const entity_inst_t& inst, bool lazy=false);
   void send_incremental_map(epoch_t since, Connection *con);
-  void send_map(MOSDMap *m, const entity_inst_t& inst, bool lazy);
   void send_map(MOSDMap *m, Connection *con);
 
 protected:
@@ -882,7 +884,7 @@ protected:
 
   void wake_pg_waiters(pg_t pgid) {
     if (waiting_for_pg.count(pgid)) {
-      take_waiters(waiting_for_pg[pgid]);
+      take_waiters_front(waiting_for_pg[pgid]);
       waiting_for_pg.erase(pgid);
     }
   }
@@ -890,7 +892,7 @@ protected:
     for (map<pg_t, list<OpRequestRef> >::iterator p = waiting_for_pg.begin();
 	 p != waiting_for_pg.end();
 	 p++)
-      take_waiters(p->second);
+      take_waiters_front(p->second);
     waiting_for_pg.clear();
   }
 
@@ -1240,8 +1242,10 @@ protected:
       osd->scrub_queue.pop_front();
       return pg;
     }
-    void _process(PG *pg) {
-      pg->scrub();
+    void _process(
+      PG *pg,
+      ThreadPool::TPHandle &handle) {
+      pg->scrub(handle);
       pg->put();
     }
     void _clear() {
@@ -1325,12 +1329,14 @@ protected:
       rep_scrub_queue.pop_front();
       return msg;
     }
-    void _process(MOSDRepScrub *msg) {
+    void _process(
+      MOSDRepScrub *msg,
+      ThreadPool::TPHandle &handle) {
       osd->osd_lock.Lock();
       if (osd->_have_pg(msg->pgid)) {
 	PG *pg = osd->_lookup_lock_pg(msg->pgid);
 	osd->osd_lock.Unlock();
-	pg->replica_scrub(msg);
+	pg->replica_scrub(msg, handle);
 	msg->put();
 	pg->unlock();
       } else {
