@@ -686,28 +686,25 @@ int RGWRados::list_objects(rgw_bucket& bucket, int max, string& prefix, string&
 			   bool get_content_type, string& ns, bool *is_truncated, RGWAccessListFilter *filter)
 {
   int count = 0;
-  string cur_marker = marker;
   bool truncated;
-  string ns_prefix;
 
   if (bucket_is_system(bucket)) {
     return -EINVAL;
   }
   result.clear();
-  if (!ns.empty()) {
-    ns_prefix = "_";
-    ns_prefix += ns + "_";
-    if (cur_marker < ns_prefix) {
-      cur_marker = ns_prefix;
-    } else if (cur_marker.substr(0, ns.size()) > ns_prefix) {
-      truncated = false;
-      goto done;
-    }
-  }
+
+  rgw_obj marker_obj, prefix_obj;
+  marker_obj.set_ns(ns);
+  marker_obj.set_obj(marker);
+  string cur_marker = marker_obj.object;
+
+  prefix_obj.set_ns(ns);
+  prefix_obj.set_obj(prefix);
+  string cur_prefix = prefix_obj.object;
 
   do {
     std::map<string, RGWObjEnt> ent_map;
-    int r = cls_bucket_list(bucket, cur_marker, prefix, max - count, ent_map,
+    int r = cls_bucket_list(bucket, cur_marker, cur_prefix, max - count, ent_map,
                             &truncated, &cur_marker);
     if (r < 0)
       return r;
@@ -1287,6 +1284,7 @@ int RGWRados::copy_obj(void *ctx,
     /* copying attrs from source, however acls should not be copied */
     attrset[RGW_ATTR_ACL] = attrs[RGW_ATTR_ACL];
   }
+  attrset.erase(RGW_ATTR_ID_TAG);
 
   RGWObjManifest manifest;
   RGWObjState *astate = NULL;
@@ -2639,31 +2637,31 @@ int RGWRados::get_obj(void *ctx, void **handle, rgw_obj& obj,
 
   state->io_ctx.locator_set_key(key);
 
+  read_len = len;
+
   if (reading_from_head) {
     /* only when reading from the head object do we need to do the atomic test */
     r = append_atomic_test(rctx, read_obj, op, &astate);
     if (r < 0)
       goto done_ret;
-  }
 
-  read_len = len;
+    if (astate) {
+      if (!ofs && astate->data.length() >= len) {
+        bl = astate->data;
+        goto done;
+      }
 
-  if (astate) {
-    if (!ofs && astate->data.length() >= len) {
-      bl = astate->data;
-      goto done;
-    }
+      if (ofs < astate->data.length()) {
+        unsigned copy_len = min((uint64_t)astate->data.length() - ofs, len);
+        astate->data.copy(ofs, copy_len, bl);
+        read_len -= copy_len;
+        read_ofs += copy_len;
+        if (!read_len)
+	  goto done;
 
-    if (ofs < astate->data.length()) {
-      unsigned copy_len = min((uint64_t)astate->data.length(), len);
-      astate->data.copy(ofs, copy_len, bl);
-      read_len -= copy_len;
-      read_ofs += copy_len;
-      if (!read_len)
-	goto done;
-
-      merge_bl = true;
-      pbl = &read_bl;
+        merge_bl = true;
+        pbl = &read_bl;
+      }
     }
   }
 
